(ns actuarial.core-test
  (:require [clojure.test :refer :all]
            [actuarial.core :refer :all]
            [actuarial.distributions :as dist]
            [actuarial.mixed-distributions :as mix]))
;; # Discrete distributions

;; ## Arithmetic distribution

(defmacro dbg [& body]
  `(let [x# ~@body]
     (println (str "dbg: " (quote ~@body) "=" x#))
     x#))

(defn close? [tolerance x y]
  (< (- x y) tolerance))

(defn =? [x y] (partial close? 0.000000001))

(deftest arith-dist
  (let [arith1 (dist/arithmetic-dist 1 0.2 0.3 0.4 0.1)]
    (testing "Arithmetic Distribution"
           (is (=? (dist/pmf arith1 3) 0.1))
           (is (=? (dist/cdf arith1 2) 0.9))
           (is (=? (dist/survival arith1 2) 0.1))
           (is (=? (dist/mean arith1) 1.4))  
           (is (=? (dist/variance arith1) 1.4))
           (is (=? (dist/value-at-risk arith1 0.6) 2))
           (is (=? (dist/tail-value-at-risk arith1 0.6) 2.25))
           )))

;; ## Poisson distribution

(deftest poisson-dist
  (let [poisson1 (dist/poisson-dist 3)
        =? (partial close? 0.00001)]
  (testing "Poisson distribution"
           (is (=? (dist/pmf poisson1 0) (Math/exp -3)))
           (is (=? (dist/mean poisson1) 3))
           (is (=? (dist/variance poisson1) 3))+
           (is (=? (dist/pgf poisson1 2) (Math/exp 3)))
           (is (=? (dist/mgf poisson1 2) (Math/exp(* 3 (- (Math/exp 2) 1)))))
           )))

;; ## Mixed Distributions

;; tests

(def dist1 (dist/arithmetic-dist 1 0.4 0.6))
(def dist2 (map #(apply dist/exponential-dist %) '([0.4] [0.5])))
(def mix1 (mix/mixedDist dist1 dist2))
(println 
  (dist/pdf mix1 0.25)
  (dist/cdf mix1 0.25)
  (dist/mgf mix1 0.1)
  (dist/rawmoments mix1 3))

(dist/describe mix1 0.25 0.5 2 0.99)

;; # Continuous distributions

;; Uniform distribution
(def uniform1 (dist/uniform-dist 1.0 3.0))
(println (dist/describe uniform1 2.0 2.5 2.0 [0.95]))

;; Exponential distribution
(def exp1 (dist/exponential-dist 2.0))
(println (dist/describe exp1 2.0 2.5 1.0 [0.95]))

;; Beta distribution
(def beta1 (dist/beta-dist 1 3))
(println (dist/describe beta1 0.75 0.25 2 [0.95]))

;; Gamma distribution
(def gamma1 (dist/gamma-dist 2.001 4.999))
(println (dist/describe gamma1 0.75 0.25 2 [0.95]))

;; Erlang distribution
(def erlang1 (dist/erlang-dist 2 5))
(println (dist/describe erlang1 0.75 0.25 2 [0.95]))

;; Weibull distribution
(def weibull1 (dist/weibull-dist 2.2 3.1))
(println (dist/describe weibull1 0.75 0.25 2 [0.95]))

;; Generalized Erlang distribution
(def generlang1 (dist/generalized-erlang-dist 2.1 5 0.25))
(println (dist/describe generlang1 0.75 0.25 2 [0.95]))

;; Gaussian distribution
(def gaussian1 (dist/gaussian-dist 1.5 0.5))
(println (dist/describe gaussian1 0.75 0.25 2 [0.95]))

;;Lognormal distribution
(def lognormal1 (dist/lognormal-dist 1.5 0.5))
(println (dist/describe lognormal1 0.75 0.25 2 [0.95]))

;;Inverse Gaussian distribution
(def inversegaussian1 (dist/inverse-gaussian-dist 1.5 0.5))
(println (dist/describe inversegaussian1 0.75 0.25 2 [0.95]))

;;Pareto distribution
(def pareto1 (dist/pareto-dist 4 3))
(println (dist/describe pareto1 0.75 0.25 2 [0.95]))

;;Generalized Pareto distribution
(def generalizedpareto1 (dist/generalized-pareto-dist 4 3 1))
(println (dist/describe generalizedpareto1 0.75 0.25 2 [0.95]))

;;Burr distribution
(def burr1 (dist/burr-dist 4 3 1))
(println (dist/describe burr1 0.75 0.25 2 [0.95]))

;;Log-Logistic distribution
(def loglog1 (dist/log-logistic-dist 3 2.1))
(println (dist/describe loglog1 0.75 0.25 2 [0.95]))

;;Student distribution
(def student1 (dist/student-dist 3))
(println (dist/describe student1 0.75 0.25 0.5 [0 1]))

(run-tests)

